{-# LANGUAGE DataKinds #-}
{-# LANGUAGE KindSignatures #-}
{-# LANGUAGE MultiParamTypeClasses #-}
{-# LANGUAGE ScopedTypeVariables #-}
{-# LANGUAGE TemplateHaskell #-}
{-# LANGUAGE TypeOperators #-}

module Agent.Tools.Tool where

import Control.Monad.IO.Class
import qualified Data.Kind as DK (Type)
import Data.Proxy
import Data.Text
import Language.Haskell.TH

data Tool = Tool
  { name :: Text,
    description :: Text,
    call :: Text
  }
  deriving (Show)

toMarkdownDL :: Tool -> Text
toMarkdownDL t = "- " <> name t <> "(" <> call t <> ")\n: " <> description t

-- | This type represents unknown type-level tools.
data SomeTool = forall t. (KnownTool t) => SomeTool (Proxy t)

-- | This class gives the tool associated with a type-level tool.
-- There are instances of the class for every concrete tool included.
class KnownTool (t :: DK.Type) where
  toolVal :: Proxy t -> Tool

class (KnownTool t) => Dispatchable t a where
  dispatch :: (MonadIO m) => Proxy t -> a -> m Text
  parseArgs :: Proxy t -> Text -> a

getToolListTH :: Q Exp
getToolListTH = do
  ClassI _ instances <- reify ''Dispatchable
  let toolExps =
        [ [|toolVal (Proxy :: Proxy $(return t))|]
        | InstanceD _ _ (AppT (AppT _ t) _) _ <-
            instances
        ]
  listE toolExps
